from ngSkinTools2.python_compatibility import Object

from PySide2 import QtWidgets, QtGui
from PySide2.QtCore import Qt

from ngSkinTools2 import cleanup, signal, licenseClient
from ngSkinTools2.observableValue import ObservableValue
from ngSkinTools2.ui import qt, dialogs
from ngSkinTools2.ui.session import withSession, session


class Message(Object):
    type_error = 'error'
    type_info = 'info'

    def __init__(self, text, type=None):
        self.type = type if type is not None else self.type_info
        self.text = text


class LicenseWindowModel(Object):
    mode_license_key = "License key"
    mode_activation_code = "Activation code"
    mode_license_server = "License server"

    def __init__(self, license_info):
        self.license_status = ObservableValue()
        self.message = ObservableValue()
        self.operation_in_progress = ObservableValue()
        self.activation_code_link = ObservableValue()

        self.__license_info = license_info
        self.license_key_activation_method = 'online'

        self.license_key = ''
        self.activation_method_selected = 'License key'

        self.activation_code = ''

        self.server_address = ''

        self.update_status()

    def update_status(self):
        self.license_status.set(self.__license_info.current_status())

    def activate(self):
        def do_license_key_online():
            self.operation_in_progress.set(True)
            task = self.__license_info.activate_with_license_key(self.license_key)

            def done(context):
                self.operation_in_progress.set(False)
                if context.error:
                    self.message.set(Message(context.error, type=Message.type_error))
                else:
                    self.message.set(None)

                self.update_status()

                if self.message() is None:
                    self.message.set(
                        Message(self.license_status().status_description, type=Message.type_info if self.license_is_active() else Message.type_error)
                    )

            task.add_done_handler(done)
            task.start()

        def do_license_key_offline():
            try:
                self.activation_code_link.set(self.__license_info.generate_acivation_code_link(self.license_key))
                self.message.set(Message("Activation code generated", Message.type_info))
            except Exception as err:
                self.message.set(Message(str(err), type=Message.type_error))

        def do_activation_code():
            try:
                self.__license_info.accept_activation_code(self.activation_code)
                self.update_status()
                self.message.set(Message("Activation code accepted", Message.type_info))
            except Exception as err:
                self.message.set(Message(str(err), type=Message.type_error))

        def do_license_server():
            self.__license_info.accept_license_server_url(self.server_address)
            self.update_status()

            self.message.set(Message("Connected to license server", Message.type_info))

        if self.activation_method_selected == self.mode_license_key:
            if self.license_key_activation_method == 'online':
                do_license_key_online()
            if self.license_key_activation_method == 'offline':
                do_license_key_offline()

        if self.activation_method_selected == self.mode_activation_code:
            do_activation_code()

        if self.activation_method_selected == self.mode_license_server:
            do_license_server()

    def clear_license_configuration(self):
        self.__license_info.clear_configuration()
        self.update_status()
        self.message.set(Message("License configuration cleared", Message.type_info))

    def license_is_active(self):
        status = self.license_status(None)
        return status is not None and status.active


@withSession
def show(parent, license_info=None):
    """
    :type license_info: object
    :type parent: QWidget
    """

    if license_info is None:
        license_info = session.licenseClient

    model = LicenseWindowModel(license_info)

    window = QtWidgets.QDialog(parent)
    window.setModal(False)
    window.resize(600, 500)
    window.setAttribute(Qt.WA_DeleteOnClose)
    window.setWindowTitle("ngSkinTools license activation")
    layout = QtWidgets.QVBoxLayout()
    window.setLayout(layout)
    layout.setMargin(0)

    # noinspection PyShadowingNames
    def license_status():
        result = QtWidgets.QWidget()
        result.setPalette(qt.alternative_palette_light())
        result.setAutoFillBackground(True)

        layout = QtWidgets.QVBoxLayout()
        result.setLayout(layout)

        layout.addWidget(QtWidgets.QLabel("<b>License status:</b>"))
        status_label = QtWidgets.QLabel("Licensed to whoever whatever (debug)")
        layout.addWidget(status_label)

        @signal.on(model.license_status.changed)
        def update_license_status():
            if model.license_is_active():
                status_label.setText("Licensed to " + str(model.license_status().licensed_to))
            else:
                status_label.setText("Evaluation/non-commercial use")

        message_label = QtWidgets.QLabel()
        layout.addWidget(message_label)
        message_label.setVisible(False)

        @signal.on(model.message.changed)
        def message_changed():
            msg = model.message()
            message_label.setVisible(msg is not None and msg != "")
            if not msg:
                return

            color = {Message.type_error: '#FF0000', Message.type_info: '#5cb85c'}[msg.type]
            message_label.setText('<b><font color="{color}">{msg}</font></b>'.format(color=color, msg=msg.text))

        update_license_status()

        layout.setMargin(25)
        return result

    def activation_method():
        result = QtWidgets.QVBoxLayout()
        result.setContentsMargins(25, 10, 25, 10)

        radio_row = QtWidgets.QHBoxLayout()
        button_group = QtWidgets.QButtonGroup()
        for index, i in enumerate(
            [LicenseWindowModel.mode_license_key, LicenseWindowModel.mode_activation_code, LicenseWindowModel.mode_license_server]
        ):
            radio = QtWidgets.QRadioButton(i)
            button_group.addButton(radio, index)
            radio_row.addWidget(radio)
        radio_row.addStretch()
        button_group.buttons()[0].setChecked(True)

        result.addWidget(QtWidgets.QLabel("Activate new license with:"))
        result.addLayout(radio_row)

        result.addSpacing(10)

        methods = [qt.wrapLayoutIntoWidget(i) for i in (license_key_activation_method(), activation_code_method(), license_server_method())]
        for i in methods:
            result.addWidget(i)

        loading_indicator = make_loading_indicator()
        result.addWidget(loading_indicator)

        @qt.on(button_group.buttonClicked)
        @signal.on(model.operation_in_progress.changed)
        def update_controls_state():
            for index, method in enumerate(methods):
                method.setVisible(not model.operation_in_progress(False) and button_group.checkedId() == index)
                method.setEnabled(not model.operation_in_progress(False))

            loading_indicator.setVisible(model.operation_in_progress(False))
            for i in button_group.buttons():
                i.setEnabled(not model.operation_in_progress(False))

            model.activation_method_selected = button_group.checkedButton().text()

        update_controls_state()

        result_widget = qt.wrapLayoutIntoWidget(result)
        return result_widget

    def make_loading_indicator():
        result = QtWidgets.QLabel("Please wait...")
        return result

    def activation_code_link_info():
        result = QtWidgets.QVBoxLayout()
        result.setMargin(0)
        result.addSpacing(10)
        result.addWidget(QtWidgets.QLabel("Use this link to generate activation code:"))
        linkText = QtWidgets.QTextEdit()
        linkText.setWordWrapMode(QtGui.QTextOption.WrapAnywhere)
        linkText.setReadOnly(True)
        linkText.setMaximumSize(5000, 70)
        result.addWidget(linkText)

        resultWidget = qt.wrapLayoutIntoWidget(result)

        @signal.on(model.activation_code_link.changed)
        def updateLicenseInfo():
            link = model.activation_code_link("")
            resultWidget.setVisible(bool(link))
            linkText.setText(link)

        updateLicenseInfo()

        copyButton = QtWidgets.QPushButton("Copy to clipboard")

        @qt.on(copyButton.clicked)
        def copyClicked():
            QtWidgets.QApplication.clipboard().setText(linkText.toPlainText())

        button_row = QtWidgets.QHBoxLayout()
        button_row.addStretch()
        button_row.addWidget(copyButton)
        result.addLayout(button_row)

        return resultWidget

    def license_key_activation_method():
        result = QtWidgets.QVBoxLayout()
        result.addWidget(QtWidgets.QLabel("License key:"))
        licenseKey = QtWidgets.QLineEdit()
        licenseKey.setPlaceholderText("00000000-0000-0000-0000-000000000000")
        result.addWidget(licenseKey)

        result.addSpacing(10)

        result.addWidget(QtWidgets.QLabel("Activation method:"))
        radio_row = QtWidgets.QVBoxLayout()
        button_group = QtWidgets.QButtonGroup()
        result.addLayout(radio_row)

        method_online = QtWidgets.QRadioButton("Connect and download license data (requires Internet connectivity)")
        method_offline = QtWidgets.QRadioButton("Offline: generate a link for downloading activation code")
        button_group.addButton(method_online)
        button_group.addButton(method_offline)
        radio_row.addWidget(method_online)
        radio_row.addWidget(method_offline)
        method_online.setChecked(True)

        result.addWidget(activation_code_link_info())

        @qt.on(method_online.toggled, method_offline.toggled)
        def update_activation_method():
            model.license_key_activation_method = 'online' if method_online.isChecked() else "offline"

        @qt.on(licenseKey.textChanged)
        def update_license_key():
            model.license_key = licenseKey.text()

        return result

    def activation_code_method():
        result = QtWidgets.QVBoxLayout()
        result.addWidget(QtWidgets.QLabel("Activation code:"))

        activationCode = QtWidgets.QTextEdit()
        activationCode.setWordWrapMode(QtGui.QTextOption.WrapAnywhere)
        activationCode.setPlaceholderText("Paste or drop activation code here.")

        @qt.on(activationCode.textChanged)
        def activationCodeUpdated():
            model.activation_code = activationCode.toPlainText()

        result.addWidget(activationCode)
        return result

    def license_server_method():
        result = QtWidgets.QVBoxLayout()
        result.addWidget(QtWidgets.QLabel("Server address:"))
        serverUrl = QtWidgets.QLineEdit()
        serverUrl.setPlaceholderText("your-server:8050 or 10.45.11.85:8050")
        result.addWidget(serverUrl)

        @qt.on(serverUrl.textChanged)
        def activationCodeUpdated():
            model.server_address = serverUrl.text()

        result.addStretch()
        return result

    # noinspection PyShadowingNames
    def buttons_row():
        btnActivate = QtWidgets.QPushButton("Activate")
        btnActivate.setDefault(True)
        btnActivate.setMinimumWidth(100)

        qt.on(btnActivate.clicked)(model.activate)

        btnClose = QtWidgets.QPushButton("Cancel")
        btnClose.setMinimumWidth(100)
        btnClose.clicked.connect(lambda: window.close())

        btnRelease = QtWidgets.QPushButton("Clear license configuration")

        @qt.on(btnRelease.clicked)
        def handle_release():
            message = (
                "This will reset license configuration so you can choose another activation method.\n"
                "For standalone licenses, this does not release license key binding to this workstation.\n"
                "\n"
                "Do you want to proceed?"
            )

            if dialogs.yesNo(message):
                model.clear_license_configuration()

        layout = QtWidgets.QHBoxLayout()
        layout.addWidget(btnRelease)
        layout.addStretch()
        layout.setContentsMargins(20, 15, 20, 15)
        layout.addWidget(btnActivate)
        layout.addWidget(btnClose)

        @signal.on(model.license_status.changed, model.operation_in_progress.changed)
        def update_license_status():
            btnActivate.setVisible(not model.license_is_active())
            btnClose.setText("Close" if model.license_is_active() else "Cancel")
            btnRelease.setVisible(model.license_is_active())

            btnActivate.setEnabled(not model.operation_in_progress(False))

        update_license_status()

        return layout

    layout.addWidget(license_status())

    method_ui = activation_method()

    @signal.on(model.license_status.changed)
    def update_controls_state():
        method_ui.setVisible(not model.license_is_active())

    layout.addWidget(method_ui)
    layout.addStretch(2)
    layout.addLayout(buttons_row())

    update_controls_state()
    window.show()

    cleanup.registerCleanupHandler(window.close)
